// -*- C++ -*-

#pragma once


#include "./defs.h"
//#include <memory>
#include <iostream>

inline void goopax_debugfunc()
{
}


namespace goopax {
namespace impl {
enum errno_type
{
    SUCCESS = 0,
    ERR_general,
    ERR_integrity,
    ERR_cl,
    ERR_cuda,
    ERR_gl,
    ERR_syntax,
    ERR_memory,
    ERR_assertion,
    ERR_math,
    ERR_io,
    ERR_too_many_regs,
    ERR_env,
    ERR_image_format
};

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.0.0) const char* goopax_ex_what2() noexcept;
GOOPAX_DLL_EXPORT(3.0.0) void goopax_io_warn3(const char*, unsigned int) noexcept;
}
}


struct EX
{
    /// base exception thrown by goopax
    struct goopax_exception : public std::exception
    {
        virtual const char* what() const noexcept final
        {
            return interface::goopax_ex_what2();
        }
        goopax_exception() = default;
    };


    struct integrity : public goopax_exception
    {
    };

    struct cl : public goopax_exception
    {
    };

    struct cuda : public goopax_exception
    {
    };

    struct gl : public goopax_exception
    {
    };

    struct syntax : public goopax_exception
    {
    };

    struct memory : public goopax_exception
    {
    };

    struct assertion : public goopax_exception
    {
    };

    struct math : public goopax_exception
    {
    };

    struct io : public goopax_exception
    {
    };

    struct too_many_regs : public goopax_exception
    {
    };

    struct env : public goopax_exception
    {};

    struct image_format : public goopax_exception
    {
    };


    template<class = void>
    static void throw_exception(errno_type err)
    {
#ifdef _WIN32
        std::cerr << "GOOPAX: Got exception '" << interface::goopax_ex_what2() << "'" << std::endl;
#endif

#define MAKEEX(ex, num) \
    if (err == num) { throw EX::ex(); }

        MAKEEX(goopax_exception, ERR_general)
        MAKEEX(integrity, ERR_integrity)
        MAKEEX(cl, ERR_cl)
        MAKEEX(cuda, ERR_cuda)
        MAKEEX(gl, ERR_gl)
        MAKEEX(syntax, ERR_syntax)
        MAKEEX(memory, ERR_memory)
        MAKEEX(assertion, ERR_assertion)
        MAKEEX(math, ERR_math)
        MAKEEX(io, ERR_io)
        MAKEEX(too_many_regs, ERR_too_many_regs)
        MAKEEX(env, ERR_env)
        MAKEEX(image_format, ERR_image_format)

#undef MAKEEX

        std::cerr << "Unknown errno=" << err << "!" << std::endl;
        abort();
    }
};

inline void check_errno(errno_type err)
{
    if (GOOPAX_EXPECT(err, SUCCESS) != SUCCESS)
    {
        EX::throw_exception(err);
    }
}

template<class R, class... A, class... B>
R if_call(errno_type (*func)(R*, A...) noexcept, B... a)
{
    R r;
    check_errno((*func)(&r, a...));
    return r;
}
}
using impl::EX;
}
