// -*- C++ -*-

#pragma once

#include "gpu-utils-core.h"
#include "struct_types.h"
#include <iostream>


namespace goopax {
namespace impl {
struct resource_impl;
template<typename T, typename SIZE_TYPE = goopax::impl::Tuint>
class buffer;

namespace gpu_struct {
struct sitem_ti_base;

#ifdef _MSC_VER
using goopax::impl::Tbool;
using goopax::impl::Tint;
using goopax::impl::Tsize_t;
using goopax::impl::Tuint;
#endif

}

template<class T>
struct opengl_texture;
namespace gpu_struct {
template<class X>
struct struct_type;
}
}
template<class STR, class U>
struct goopax_struct_changetype<STR, U, typename impl::enable_if<impl::is_arithmetic<STR>::value || impl::is_pointer<STR>::value>::type>
{
    using type = typename impl::gpu_struct::struct_type<U>::template type<STR>;
};

template<class T, class U>
struct goopax_struct_changetype<gpu_type<T>, U>
{
    using type = typename impl::gpu_struct::struct_type<U>::template type<T>;
};


template<typename T, typename TO, typename U>
struct goopax_struct_changetype<impl::gpu_struct::struct_change_type<T, TO>, U>
{
    using type = typename goopax_struct_changetype<TO, U>::type;
};


namespace impl {
namespace gpu_struct {
struct mapframe_base;

struct sitem_ti_base
{
    virtual void out(ostream& s, const void* value) const = 0;
};

template<typename T>
void sitem_out(ostream& s, const T& v)
{
    if (is_uninitialized(v))
    {
        s << v;
    }
    else
    {
        s << unrange(v) + 0;
    }
}
template<typename T>
void sitem_out(ostream& s, const T* v)
{
    s << (const void*)v;
}
template<typename T>
void sitem_out(ostream&, const gpu_type<T>&)
{
    abort();
}

template<class T>
struct sitem_ti : public sitem_ti_base
{
    static const sitem_ti info;

    virtual void out(ostream& s, const void* value) const final override
    {
        sitem_out(s, *static_cast<const T*>(value));
    }
};
template<class T>
const sitem_ti<T> sitem_ti<T>::info;


struct mapframe_entry
{
    Tuint                offset;
    Tuint                type;
    const sitem_ti_base* si = nullptr;

    mapframe_entry(Tuint offset0, Tuint type0, const sitem_ti_base* si0)
        : offset(offset0)
        , type(type0)
        , si(si0)
    {}

    mapframe_entry() {}
};
}

namespace interface {
extern "C" {
GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_sitem_set_memref2(gpu_writer_base_impl&, const gpu_writer_base_impl&, Uint, resource_impl& base, Uint elesize, bool) noexcept;
GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.1.0) void goopax_sitem_set_memref3(gpu_writer_base_impl&, const gpu_writer_base_impl&, Uint, resource_impl* base, Uint elesize, bool, bool) noexcept;
GOOPAX_DLL_EXPORT(3.1.1) void goopax_sitem_set_memref4(const gpu_struct::mapframe_base& map, const gpu_struct::mapframe_base& gpu_map, const gpu_writer_base_impl& pos, resource_impl* base, Tuint elesize, void* ref) noexcept;
GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_cleardebug2(goopax_device_impl* device) noexcept;
GOOPAX_DLL_EXPORT(3.1.1) void goopax_struct_cond(const gpu_struct::mapframe_base& gpu_map, const gpu_writer_base_impl*, const void*, const void*, void*);
}
}

namespace gpu_struct {
template<class X>
struct struct_type;

template<typename STR, Size_t N>
struct modtest;


template<typename = void>
struct sitem_info
{
    static mutex& Mutex()
    {
        static mutex M;
        return M;
    }
    static mapframe_entry* p;
    static const void*     struct_begin;
    static const void*     struct_end;
};
template<typename T>
mapframe_entry* sitem_info<T>::p;
template<typename T>
const void* sitem_info<T>::struct_begin;
template<typename T>
const void* sitem_info<T>::struct_end;


struct struct_item_map_base
{
    struct_item_map_base()                            = default;
    struct_item_map_base(const struct_item_map_base&) = delete;
};


template<class T>
struct struct_item_map : public struct_item_map_base
{
    using TEST = typename gettype<T>::cpu;
    //	  static_assert(is_fundamental<TEST>::value || is_half<TEST>::value, "");
    T data;

    struct_item_map(const struct_item_map&)
        : struct_item_map()
    {}

    struct_item_map();

    template<class U>
    struct_item_map(const U&)
        : struct_item_map()
    {}

    ~struct_item_map()
    {
    }
};


template<typename T, Size_t N>
struct struct_mod_test
{
    array<Size_t, 1 + (gettype<T>::size % N)> data;
};
template<typename T, Uint POW>
struct struct_type_probe
{
    array<Size_t, 1 + pow<POW, 1>(type_enum<T>::value)> data;
};

template<typename T, typename TO>
struct struct_change_type
{
};


struct mapframe_base
{
    const mapframe_entry* tab_p;
    Tuint                 N;
    const Tuint           structsize;
};

struct init_struct
{
    init_struct(void* begin, void* end)
    {
        sitem_info<>::struct_begin = begin;
        sitem_info<>::struct_end   = end;
    }
};


template<class STR>
struct mapframe : public mapframe_base
{
    array<mapframe_entry, modtest<STR, 1>::value> tab;

    static const mapframe& get_static_mf()
    {
        static const mapframe static_mf;
        return static_mf;
    }

    mapframe()
        : mapframe_base({ nullptr, modtest<STR, 1>::value, sizeof(STR) })
    {
        auto&             M = sitem_info<>::Mutex();
        lock_guard<mutex> Lock(M);
        sitem_info<>::p = tab.data();


        struct wrap
        {
            init_struct                                                                                   init;
            typename goopax_struct_changetype<STR, struct_item_map<typename get_X_type<STR>::type>>::type str;

            wrap()
                : init(&str, &str + 1)
                , str()
            {
                static_assert(sizeof(STR) == sizeof(str), "GOOPAX: something wrong with gpu struct!");
            }
        } Wrap;

        assert(sitem_info<>::p <= tab.data() + tab.size());
        this->N     = sitem_info<>::p - tab.data();
        this->tab_p = tab.data();
    }
};


template<class T>
void unspecified_dump2(ostream& s, const T& v, false_type, false_type)
{
    s << v;
}
template<class T>
void unspecified_dump2(ostream& s, const T& v, false_type, true_type)
{
    s << (Tint)v;
}
template<class T, class B>
void unspecified_dump2(ostream& s, const T& v, true_type, B)
{
    (void)v;
    s << "[gpu_type]";
}


/*      inline void makeref(const mapframe_base& map, const mapframe_base& gpu_map, const gpu_writer_base_impl& pos, resource_impl* base, Tuint elesize, void* ref)
      {
	  interface::goopax_sitem_set_memref4(map, gpu_map, pos, base, elesize, ref);
      }
*/

/// \ingroup memory
/// base for resource class.
template<class STR, class BASE>
struct sresource
{
    using BASE_t = BASE;
    BASE base;

    using value_type      = typename goopax_struct_changetype<STR, xtype_gpu>::type;
    using reference       = value_type&;
    using const_reference = const value_type&;
    using size_type       = typename BASE::size_type;
    using ssize_type      = typename BASE::ssize_type;
    using maptype         = mapframe<typename goopax_struct_changetype<STR, xtype_int>::type>;
    using gpu_maptype     = mapframe<typename goopax_struct_changetype<STR, xtype_gpu>::type>;

    size_type size() const
    {
        return base.size() * gettype<typename BASE::value_type>::size / gettype<value_type>::size;
    }

    private:
    reference makeref(const size_type& pos) const
    {
        const size_type basepos = pos * (gettype<STR>::size / gettype<typename BASE::value_type>::size);
        static_assert(gettype<STR>::size % gettype<typename BASE::value_type>::size == 0, "");
        value_type* ret = new value_type;
        interface::goopax_register_garbage(ret, garbage_deleter<value_type>);
        interface::goopax_sitem_set_memref4(maptype::get_static_mf(), gpu_maptype::get_static_mf(), *basepos.impl, base.impl.get(), gettype<typename BASE::value_type>::size, ret);
        return *ret;
    }

    public:
    reference operator[](const size_type& pos)
    {
        /// element access
        return makeref(pos);
    }

    const_reference operator[](const size_type& pos) const
    {
        /// element access
        return makeref(pos);
    }

    void barrier() const
    {
        base.barrier();
    }

    sresource()
    {}

    template<class... ARGS>
    sresource(const ARGS&... args)
        : base(args...)
    {
        static_assert(gettype<STR>::size % gettype<typename BASE::value_type>::size == 0, "");
    }
};


template<class X, class T>
struct get_struct_type;

template<class T, class U>
struct get_struct_type<struct_item_map<U>, T>
{
    using type = struct_item_map<typename gettype<U>::template change<T>::type>;
};
template<class T>
struct get_struct_type<struct_item_map<xtype_int>, T>
{
    using type = struct_item_map<typename gettype<Uint>::template change<T>::type>;
};

template<class T>
struct get_struct_type<xtype_gpu, T>
{
    using type = typename gettype<T>::gpu;
};
template<class T>
struct get_struct_type<xtype_int, T>
{
    using type = typename unrangetype<T>::type;
};
template<class T, class TA>
struct get_struct_type<checknan<xtype_int, TA>, T>
{
    using type = checknan<typename unrangetype_core<T>::type, TA>;
};

template<typename T, typename U, Size_t N>
struct get_struct_type<struct_mod_test<U, N>, T>
{
    using type = struct_mod_test<typename gettype<U>::template change<T>::type, N>;
};
template<typename T, Size_t N>
struct get_struct_type<struct_mod_test<xtype_int, N>, T>
{
    using type = struct_mod_test<typename gettype<Uint>::template change<T>::type, N>;
};
template<typename T, typename U, Uint POW>
struct get_struct_type<struct_type_probe<U, POW>, T>
{
    using type = struct_type_probe<typename gettype<U>::template change<T>::type, POW>;
};
template<typename T, Uint POW>
struct get_struct_type<struct_type_probe<xtype_int, POW>, T>
{
    using type = struct_type_probe<typename gettype<Uint>::template change<T>::type, POW>;
};
template<typename T, typename U, typename TO>
struct get_struct_type<struct_change_type<U, TO>, T>
{
    using type = struct_change_type<typename gettype<U>::template change<T>::type, TO>;
};
template<typename T, typename TO>
struct get_struct_type<struct_change_type<xtype_int, TO>, T>
{
    using type = struct_change_type<typename gettype<Uint>::template change<T>::type, TO>;
};


template<Size_t N>
struct basetype_t
{
    using type = typename conditional<sizeof(unsigned char) == N, unsigned char,
                                      typename conditional<sizeof(unsigned short) == N, unsigned short,
                                                           typename conditional<sizeof(unsigned int) == N, unsigned int,
                                                                                typename conditional<sizeof(unsigned long) == N, unsigned long,
                                                                                                     unsigned long long>::type>::type>::type>::type;
    static_assert(sizeof(type) == N, "");
};

template<typename STR, Size_t N>
struct modtest : public integral_constant<Size_t, sizeof(typename goopax_struct_changetype<typename remove_cv<STR>::type, struct_mod_test<xtype_int, N>>::type) / sizeof(Size_t)>
{};

template<typename STR>
struct type_probe
{
    static const Size_t N = modtest<STR, 1>::value;

    private:
    static const Size_t type_enum1 = sizeof(typename goopax_struct_changetype<typename remove_cv<STR>::type, struct_type_probe<xtype_int, 1>>::type) / sizeof(Size_t) - N;
    static const Size_t type_enum2 = sizeof(typename goopax_struct_changetype<typename remove_cv<STR>::type, struct_type_probe<xtype_int, 2>>::type) / sizeof(Size_t) - N;
    static_assert(type_enum2 * N == pow2(type_enum1), "Different types used in struct");
    using Utype = typename unrangetype_core<typename reverse_type_enum<type_enum1 / N>::type>::type;

    public:
    using type = typename conditional<is_same<STR, typename unrangetype<STR>::type>::value, Utype, checknan<Utype>>::type;
};


template<typename STR, Size_t N = 1, Bool finished = (modtest<STR, N * 2>::value != modtest<STR, 1>::value || (modtest<STR, N * 2>::value == 0 && N == 8))>
struct min_elesize;

template<typename STR, Size_t N>
struct min_elesize<STR, N, true>
{
    using type = typename basetype_t<N>::type;
    static_assert(sizeof(type) == N, "");
};

template<typename STR, Size_t N>
struct min_elesize<STR, N, false> : public min_elesize<STR, N * 2>
{};


template<class X>
struct structX_type
{
    template<class T>
    struct typegen
    {
        static_assert(is_arithmetic<typename goopax::impl::gettype<goopax::impl::Tuint>::template change<goopax::impl::Tuint>::type>::value, "");
        static_assert(is_arithmetic<T>::value || is_pointer<T>::value, "");
        using type = typename get_struct_type<X, T>::type;
    };

    template<class T>
    struct typegen<gpu_type<T>> : public typegen<T>
    {};

    template<class T>
    using type = typename typegen<T>::type;

    using uint8_type  = type<goopax::impl::Tuint8_t>;
    using int8_type   = type<goopax::impl::Tint8_t>;
    using char_type   = type<char>;
    using uint16_type = type<goopax::impl::Tuint16_t>;
    using int16_type  = type<goopax::impl::Tint16_t>;
    using uint32_type = type<goopax::impl::Tuint32_t>;
    using int32_type  = type<goopax::impl::Tint32_t>;
    using uint64_type = type<goopax::impl::Tuint64_t>;
    using int64_type  = type<goopax::impl::Tint64_t>;
    using float_type  = type<goopax::impl::Tfloat>;
    using double_type = type<goopax::impl::Tdouble>;
    using uint_type   = type<goopax::impl::Tuint>;
    using int_type    = type<goopax::impl::Tint>;
    using bool_type   = type<goopax::impl::Tbool>;
};

template<class T>
struct struct_type : public structX_type<typename get_X_type<T>::type>
{};

template<class... Args>
struct type_list
{
    template<std::size_t N>
    using type = typename std::tuple_element<N, std::tuple<Args...>>::type;
};

#define goopax_struct_defs(...)                                                                                 \
    using goopax_struct_type     = typename goopax::impl::gpu_struct::type_list<__VA_ARGS__>::template type<0>; \
    using goopax_struct_typedefs = goopax::impl::gpu_struct::struct_type<goopax_struct_type>;                   \
    using uint8_type             = typename goopax_struct_typedefs::uint8_type;                                 \
    using int8_type              = typename goopax_struct_typedefs::int8_type;                                  \
    using char_type              = typename goopax_struct_typedefs::char_type;                                  \
    using uint16_type            = typename goopax_struct_typedefs::uint16_type;                                \
    using int16_type             = typename goopax_struct_typedefs::int16_type;                                 \
    using uint32_type            = typename goopax_struct_typedefs::uint32_type;                                \
    using int32_type             = typename goopax_struct_typedefs::int32_type;                                 \
    using uint64_type            = typename goopax_struct_typedefs::uint64_type;                                \
    using int64_type             = typename goopax_struct_typedefs::int64_type;                                 \
    using float_type             = typename goopax_struct_typedefs::float_type;                                 \
    using double_type            = typename goopax_struct_typedefs::double_type;                                \
    using uint_type              = typename goopax_struct_typedefs::uint_type;                                  \
    using int_type               = typename goopax_struct_typedefs::int_type;                                   \
    using bool_type              = typename goopax_struct_typedefs::bool_type;

template<typename T, typename cpu_size_type, memory::scope SCOPE, typename ENABLE = void, typename F = void>
struct resource_type_switch;

template<typename T, typename cpu_size_type, memory::scope SCOPE>
struct resource_type_switch<T, cpu_size_type, SCOPE, typename enable_if<is_gpu_struct<T>::value>::type, void>
{
    static_assert(is_same<T, typename goopax_struct_changetype<T, typename goopax_struct_type<T>::type>::type>::value, "There seems to be something wrong with the gpu struct definition. goopax_struct_changetype does not match. If using multiple template arguments, make sure that all are of the same type class, either cpu intrinsic, cpu debug, or gpu");
    using basetype   = typename min_elesize<T>::type;
    using d_basetype = typename goopax_struct_changetype<basetype, typename get_X_type<T>::type>::type;
    using T_void     = typename goopax_struct_changetype<T, xtype_void>::type;
    using uav_type   = sresource<T_void, uav<d_basetype, cpu_size_type>>;
    using lds_type   = sresource<T_void, lds<d_basetype, SCOPE>>;
};

template<typename T, typename cpu_size_type, memory::scope SCOPE>
struct resource_type_switch<T, cpu_size_type, SCOPE, typename enable_if<is_arithmetic<T>::value && !is_same<Bool, typename unrangetype<T>::type>::value>::type, void>
{
    using d_basetype = T;
    using T_void     = typename goopax_struct_changetype<T, xtype_void>::type;
    using uav_type   = sresource<T_void, uav<T, cpu_size_type>>;
    using lds_type   = sresource<T_void, lds<T, SCOPE>>;
};
template<typename T, typename cpu_size_type, memory::scope SCOPE>
struct resource_type_switch<T, cpu_size_type, SCOPE, typename enable_if<is_arithmetic<T>::value && is_same<Bool, typename unrangetype<T>::type>::value>::type, void>
{
    using d_basetype = typename min_elesize<T>::type;
    using T_void     = typename goopax_struct_changetype<d_basetype, xtype_void>::type;
    using uav_type   = sresource<T_void, uav<T, cpu_size_type>>;
    using lds_type   = sresource<T_void, lds<T, SCOPE>>;
};

template<typename T, typename cpu_size_type, typename E, memory::scope SCOPE>
struct resource_type_switch<T, cpu_size_type, SCOPE, E, void>
{
    static_assert(sizeof(T) == 0, "Invalid resource type! Use intrinsic types (int, float, ...) or goopax struct types.");
};


template<typename T>
struct get_ptr_base_type;

template<typename T>
struct get_ptr_base_type
{
    using type = T;
};
template<typename T>
struct get_ptr_base_type<T*>
{
    using type = typename resource_type_switch<T, Tuint, memory::none>::d_basetype;
};


template<class T>
struct_item_map<T>::struct_item_map()
{
    if (this >= sitem_info<>::struct_begin && this < sitem_info<>::struct_end)
    {
        *sitem_info<>::p++ = mapframe_entry(static_cast<const char*>((void*)this) - static_cast<const char*>(sitem_info<>::struct_begin), type_enum<typename get_ptr_base_type<typename conditional<is_gpu_type<T>::value, typename gettype<T>::cpu, T>::type>::type>::value,
                                            &sitem_ti<T>::info);
    }
}
}

using gpu_struct::mapframe;
using gpu_struct::sitem_ti_base;
using gpu_struct::struct_type;


template<class A>
A cond(gpu_bool c, const A& a, const A& b, typename enable_if<is_gpu_struct<A>::value && is_same<typename get_X_type<A>::type, xtype_gpu>::value>::type* = nullptr)
{
    /// \ingroup operators
    /// conditional operator.
    /// \param c condition
    /// \param a input value a
    /// \param b input value b
    /// \param return c ? a : b
    A ret;
    interface::goopax_struct_cond(mapframe<A>::get_static_mf(), c.impl.get(), &a, &b, &ret);

    return ret;
}


template<class A>
A cond(bool c, const A& a, const A& b, typename enable_if<is_gpu_struct<A>::value && is_checknan<typename get_X_type<A>::type>::value>::type* = nullptr)
{
    /// \ingroup operators
    /// conditional operator.
    /// \param c condition
    /// \param a input value a
    /// \param b input value b
    /// \param return c ? a : b
    if (c)
        return a;
    else
        return b;
}

template<class A>
A cond(bool c, const A& a, const A& b, typename enable_if<is_gpu_struct<A>::value && is_same<typename get_X_type<A>::type, xtype_int>::value>::type* = nullptr)
{
    /// \ingroup operators
    /// conditional operator.
    /// \param c condition
    /// \param a input value a
    /// \param b input value b
    /// \param return c ? a : b
    if (c)
        return a;
    else
        return b;
}

}


//  using impl::struct_type;
using impl::cond;
}


namespace goopax {
namespace impl {
template<typename STR>
struct output_goopax_struct_helper
{
    STR data;

    template<typename S>
    friend S& operator<<(S& s, const output_goopax_struct_helper& str)
    {
        using maptype = goopax::impl::mapframe<STR>;

        s << "[";
        for (goopax::impl::Tsize_t k = 0; k < maptype::get_static_mf().N; ++k)
        {
            if (k != 0) s << ",";
            maptype::get_static_mf().tab_p[k].si->out(s, (const char*)&str + maptype::get_static_mf().tab_p[k].offset);
        }
        s << "]";
        return s;
    }
};

template<typename T>
output_goopax_struct_helper<T> output_goopax_struct(const T& t)
{
    /// \ingroup goopax_struct
    /// Outputs the contents of a goopax struct.
    return output_goopax_struct_helper<T>({ t });
}

}
}
