// -*- C++ -*-

#pragma once


#include "gpu_types.h"

/// \addtogroup goopax_struct goopax struct types
/// here some C++ types are prepared to be used as goopax struct. Other struct types need to be prepared in a similar way before they can be used to store memory in goopax.

namespace goopax {
/// std::array and derived classes \ingroup goopax_struct
template<typename T, Size_t N>
struct goopax_struct_type<std::array<T, N>>
{
    using type = T;
};

template<typename T, Size_t N, typename X>
struct goopax_struct_changetype<std::array<T, N>, X>
{
    using type = std::array<typename goopax_struct_changetype<T, X>::type, N>;
};


#if GOOPAX_HAVE_DEBUG_STL
template<typename T, Size_t N>
struct goopax_struct_type<std::__debug::array<T, N>>
{
    using type = T;
};

template<typename T, Size_t N, typename X>
struct goopax_struct_changetype<std::__debug::array<T, N>, X>
{
    using type = std::__debug::array<typename goopax_struct_changetype<T, X>::type, N>;
};

#endif


namespace impl {

template<typename T>
struct prepare_struct_helper;

template<typename X, typename HAVETUPLE, typename TODOTUPLE>
struct struct_convert;
template<typename X, typename CUR, typename... TODO, typename... HAVE>
struct struct_convert<X, std::tuple<HAVE...>, std::tuple<CUR, TODO...>>
{
    using type = typename struct_convert<X, std::tuple<HAVE..., typename goopax_struct_changetype<CUR, X>::type>, std::tuple<TODO...>>::type;
};
template<typename X, typename... HAVE>
struct struct_convert<X, std::tuple<HAVE...>, std::tuple<>>
{
    using type = std::tuple<HAVE...>;
};


#define GOOPAX_PREPARE_STRUCT(...)                                                                                                                                             \
    namespace goopax {                                                                                                                                                         \
    template<typename... A>                                                                                                                                                    \
    struct goopax_struct_type<__VA_ARGS__<A...>>                                                                                                                               \
    {                                                                                                                                                                          \
        using type = typename std::tuple_element<0, std::tuple<A...>>::type;                                                                                                   \
    };                                                                                                                                                                         \
    namespace impl {                                                                                                                                                           \
    template<typename... B>                                                                                                                                                    \
    struct prepare_struct_helper<__VA_ARGS__<B...>>                                                                                                                            \
    {                                                                                                                                                                          \
        template<typename T>                                                                                                                                                   \
        struct temptype;                                                                                                                                                       \
        template<typename... A>                                                                                                                                                \
        struct temptype<std::tuple<A...>>                                                                                                                                      \
        {                                                                                                                                                                      \
            using type = __VA_ARGS__<A...>;                                                                                                                                    \
        };                                                                                                                                                                     \
    };                                                                                                                                                                         \
    }                                                                                                                                                                          \
    template<typename... A, typename X>                                                                                                                                        \
    struct goopax_struct_changetype<__VA_ARGS__<A...>, X>                                                                                                                      \
    {                                                                                                                                                                          \
        using type = typename impl::prepare_struct_helper<__VA_ARGS__<A...>>::template temptype<typename impl::struct_convert<X, std::tuple<>, std::tuple<A...>>::type>::type; \
    };                                                                                                                                                                         \
    }

}
}


/// \}
