// -*- C++ -*-

#pragma once
#include <complex>
#include <tuple>


#include "struct_types.h"


#if defined(__APPLE__) || defined(__ANDROID__)
// complex type from Apple's xcode needs some overloading.
namespace std {
template<typename T>
goopax::gpu_type<T> norm(complex<goopax::gpu_type<T>> a)
{
    return a.real() * a.real() + a.imag() * a.imag();
}
template<typename T>
complex<goopax::gpu_type<T>> operator*(complex<goopax::gpu_type<T>> a, complex<goopax::gpu_type<T>> b)
{
    return { a.real() * b.real() - a.imag() * b.imag(), a.real() * b.imag() + a.imag() * b.real() };
}
template<typename T>
complex<goopax::gpu_type<T>> exp(complex<goopax::gpu_type<T>> a)
{
    return exp(a.real()) * complex<goopax::gpu_type<T>>(cos(a.imag()), sin(a.imag()));
}
template<class _Tp>
complex<goopax::gpu_type<_Tp>>
polar(const goopax::gpu_type<_Tp>& __rho, const goopax::gpu_type<_Tp>& __theta = _Tp(0))
{
    goopax::gpu_type<_Tp> __x = __rho * cos(__theta);
    goopax::gpu_type<_Tp> __y = __rho * sin(__theta);
    return complex<goopax::gpu_type<_Tp>>(__x, __y);
}
}
#endif


namespace goopax {
namespace impl {
template<typename T, typename EIF>
struct unrangetype_help
{
    using type = T;
};
template<typename T>
struct unrangetype_help<T, typename enable_if<goopax::impl::is_gpu_struct<T>::value>::type>
{
    using type = typename goopax::goopax_struct_changetype<typename remove_cv<T>::type, typename unrangetype<typename goopax::goopax_struct_type<T>::type>::type>::type;
};

}
}


GOOPAX_PREPARE_STRUCT(std::complex)
GOOPAX_PREPARE_STRUCT(std::pair)
GOOPAX_PREPARE_STRUCT(std::tuple)
