// -*- C++ -*-
#pragma once

#include "pointer.h"
#include "struct.h"


namespace goopax {
namespace impl {
class svm_buffer_impl;

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(3.1.0) errno_type goopax_E_svm_create(svm_buffer_impl**, goopax_device_impl*, Size_t, void**, BUFFER_FLAGS flags) noexcept;
}
}

template<typename T>
class svm_buffer
{
    static_assert(!is_checknan<T>::value, "sorry, debug types not yet allowed for svm memory");
    unique_ptr<svm_buffer_impl, interface::goopax_deleter> impl;
    T*                                                     ptr;
    goopax::impl::Tsize_t                                  size_i;

    public:
    goopax::impl::Tsize_t size() const
    {
        return size_i;
    }

    T* data()
    {
        return ptr;
    }
    const T* data() const
    {
        return ptr;
    }

    T* begin()
    {
        return data();
    }
    const T* begin() const
    {
        return data();
    }
    T* end()
    {
        return begin() + size();
    }
    const T* end() const
    {
        return begin() + size();
    }

    T& operator[](size_t k)
    {
        return begin()[k];
    }
    const T& operator[](size_t k) const
    {
        return begin()[k];
    }

    svm_buffer(goopax_device device, Tsize_t size, BUFFER_FLAGS flags = BUFFER_READ_WRITE)
        : size_i(size)
    {
        void* ptr;
        impl.reset(if_call(interface::goopax_E_svm_create, device.get_impl(), size * sizeof(T), &ptr, flags));
        this->ptr = reinterpret<T*>(ptr);
    }
    svm_buffer(svm_buffer&&) = default;
    svm_buffer& operator=(svm_buffer&&) = default;

    ~svm_buffer()
    {
    }
};


}

using impl::svm_buffer;
}
